#define _DEFAULT_SOURCE
#include <stddef.h>
#include <unistd.h>
#include <stdio.h>

/** velikost slova */
#define WORD_SIZE       (sizeof(void *))

/** velikost hlavicky alokovaneho bloku */
#define HDR_SIZE	(WORD_SIZE * 2)

/** zaokrouhli hodnotu na velikost dvou slov */
#define ROUND_UP(x)     ((x + (HDR_SIZE - 1)) & ~(HDR_SIZE - 1))
#define MIN_ACQ_SIZE	(4096 * 16)

/** struktura popisujici alokovany blok pameti, resp. uvolneny blok */
struct chunk {
        size_t prev_size;    // velikost predchoziho bloku (v bytech, vcetne hlavicky), nejnizsi bit signalizuje, ze je dany blok volny
        size_t size;         // velikost bloku (v bytech, vcetne hlavicky)
        struct chunk *prev;  // odkaz na predchozi volny blok
        struct chunk *next;  // odkaz na dalsi volny blok
};

/** ukazatel na volnou oblast haldy */
unsigned char *data_area = NULL;

/** ukazatel na zacatek haldy */
unsigned char *data_area_start = NULL;

/** velikost haldy v bytech */
size_t data_area_capacity = 0;

/** ukazatel na seznam volnych bloku */
struct chunk *free_blocks = NULL;

/** ukazatel na posledni alokovany blok na halde */
struct chunk *last_chunk = NULL;

/** rozsiri haldu o MIN_ACQ_SIZE */
static void acquire_mem()
{
	unsigned char *last_pos = sbrk(MIN_ACQ_SIZE);
	if (!data_area_start) {
		data_area_start = last_pos;
		data_area = data_area_start;
	}
	data_area_capacity += MIN_ACQ_SIZE;
}

/** prevede odkaz na ukazatel na blok pameti (vcetne hlavicky) */
static inline struct chunk *ptr_to_chunk(void *ptr)
{
	return (struct chunk *) (((unsigned char *) ptr) - HDR_SIZE);
}

/** prevede ukazatel na blok pameti (vcetne hlavicky) na ukazatel,
  * ktery ukazuje do datove oblasti bloku (bez hlavicky) */
static inline void *chunk_to_ptr(struct chunk *mem)
{
	return ((unsigned char *) mem) + HDR_SIZE;
}

/** vraci true, pokud je dany blok uvolneny */
static inline int chunk_is_free(struct chunk *mem)
{
	return mem->prev_size & 0x01;
}

/** nastavi, jestli je dany blok pouzivany nebo volny */
static inline void chunk_set_free(struct chunk *mem, int free)
{
	mem->prev_size = (mem->prev_size & ~0x01) | free;
}

/** vraci predchozi blok na halde, pokud neexistuje, vraci NULL */
static inline struct chunk *chunk_preceding(struct chunk *mem)
{
	unsigned char *pos = (unsigned char *) mem; 
	if (pos == data_area_start) return NULL;
	return (struct chunk *) (pos - (mem->prev_size & ~0x01));
}

/** vraci nasledujici blok na halde, pokud neexistuje, vraci NULL */
static inline struct chunk *chunk_succeeding(struct chunk *mem)
{
	unsigned char *pos = (unsigned char *) mem;
	if ((pos + mem->size) == data_area) return NULL;
	return (struct chunk *) (pos + mem->size);
}


/** zaradi blok na seznam volnych bloku */
static void chunk_enqueue(struct chunk *mem)
{
	if (free_blocks) {
		free_blocks->prev = mem;;
	}

	mem->next = free_blocks;
	mem->prev = NULL;
	free_blocks = mem;

	chunk_set_free(mem, 1);
}

/** vyjme zadany blok ze seznamu volnych bloku */
static void chunk_dequeue(struct chunk *mem)
{
	if (mem == free_blocks) {
		free_blocks = free_blocks->next;
		if (free_blocks) {
			free_blocks->prev = NULL;
		}
	} else {
		mem->prev->next = mem->next;
		if (mem->next) {
			mem->next->prev = mem->prev;
		}
	}
}

/** na halde vytvori novy blok zadane velikosti (musi byt nasobek velikosti dvou slov */
static struct chunk *chunk_allocate(size_t size)
{
	while (size > data_area_capacity) {
		acquire_mem();
	}

	struct chunk *mem = (struct chunk *) data_area;
	data_area += size;
	data_area_capacity -= size;

	mem->size = size;

	if (last_chunk) {
		mem->prev_size = last_chunk->size;
	} else {
		mem->prev_size = 0;
	}
	
	last_chunk = mem;
	return mem;
}

/** zaradi uvolneny blok na seznam volnych bloku, provadi slouceni s vedlejsimi bloky */
static void chunk_enqueue_merged(struct chunk *mem)
{
	struct chunk *prec = chunk_preceding(mem);	
	struct chunk *succ = chunk_succeeding(mem);	

	if (prec && chunk_is_free(prec)) {
		chunk_dequeue(prec);
		prec->size += mem->size;
		mem = prec;
	}

	if (succ && chunk_is_free(succ)) {
		chunk_dequeue(succ);
		mem->size += succ->size;
		succ = chunk_succeeding(mem);
	}

	if (succ) { 
		succ->prev_size = mem->size;
		chunk_enqueue(mem);
	} else {
		data_area -= mem->size;
		data_area_capacity += mem->size;
		last_chunk = chunk_preceding(mem);
	}
}

/** mezi volnymi bloky najde prvni vhodne velikosti, pokud neexistuje, vraci NULL */
static struct chunk *chunk_find(size_t size)
{
	if (!free_blocks) return NULL;
	struct chunk *c = free_blocks;
	while (c) {
		// strategie first fit
		if (c->size >= size) {
			chunk_dequeue(c);
			return c;
		}
		c = c->next;
	}
	return NULL;
}

/** ekvivalent funkce malloc */
void *tmalloc(size_t size)
{
	size = ROUND_UP(size + HDR_SIZE);

	struct chunk *mem = chunk_find(size);
	if (mem) {
		chunk_set_free(mem, 0);
	} else {
		mem = chunk_allocate(size);
	}

	return chunk_to_ptr(mem);
}


/** ekvivalent funkce free */
void tfree(void *ptr)
{
	struct chunk *mem = ptr_to_chunk(ptr); 
	chunk_enqueue_merged(mem);
}

/** funkce umoznujici zobrazit informace o halde vcetne jednotlivych objektu */
void tmalloc_debug()
{
	printf("heap start:\t%p\n", data_area_start);
	printf("heap end:\t%p\n", data_area);
	printf("heap size:\t%li\n", data_area - data_area_start);
	printf("heap free capacity:\t%li\n", data_area_capacity);

	if (data_area_start == data_area) return;

	struct chunk *mem = (struct chunk *) data_area_start;
	for (; mem != NULL; mem = chunk_succeeding(mem)) {
		printf("%p\t%zi\t%zi\t%s (prev: %zi)\n",
			mem,
			mem->size - HDR_SIZE,
			mem->size,
			chunk_is_free(mem) ? "free" : "used",
			mem->prev_size);
	}
}
